import numpy as np
import taichi as ti
import random
import math
from PIL import Image
from numpy.random import default_rng
import crowdsim.utils
import json

@ti.data_oriented
class Config:
    def __init__(self):
        self.Astar = 0 # 是否使用A*预渲染地图 1代表是
        self.dynamic_search = 0 #邻域搜索时使用静态分配内存还是动态分配内存 1代表用dynamic node 0代表用静态分配内存方法
        self.analysis = 0 #是否进行分析 1代表是

        self.sim_step = 0.2  # 仿真步长
        self.WINDOW_HEIGHT = 200 # 预定义的窗口大小 如果读图片则窗口大小为图片分辨率
        self.WINDOW_WIDTH = 200 
        self.expand_size = 1
        self.window_size = 200  # 每条边划分为多少个格子 这个值调大了之后影响邻域搜索 这个值目前还代表着地图的大小（windowsize米
        self.pixel_number = self.window_size ** 2
        self.search_radius = 20 # 搜索半径 eg.search_radius=2代表在5*5的网格中进行搜索
        self.max_neighbour = 10 #agent/obstacle的值当达到max_neighbour之后，就不再计算，防止人群dense或者障碍物dense出现震荡
        self.stop_radius = 0.01 # 如果足够接近目标，stop   

        self.path = ""
        self.im = None

        # desired force
        self.relaxation_time = 0.5
        self.goal_threshold = 0.01 # 减速的threshold
        self.desired_factor = 10

        # fij
        self.social_radius = 0.01
        self.lambda_importance = 2.0
        self.gamma = 0.35
        self.n_fij = 2
        self.n_prime = 3
        self.fij_factor = 10

        # obstacle force
        self.shoulder_radius = 0.001
        self.sigma = 0.2
        self.obstacle_threshold = 0.02 #障碍物/点比较密集时 可能需要调小以取得更好的效果
        self.obstacle_factor = 10

        # social_force advanced
        self.use_gender = 0
        self.gender_gap = 1.5

        # flocking
        self.flocking = 0 #为1时才在仿真中加入flocking force
        self.flocking_radius = 0.15
        self.alignment_factor = 10
        self.separation_factor = 10
        self.cohesion_factor = 0

        # steering force
        self.steering_force = 0 #为1时才在仿真中加入steering force
        self.flee_factor = 1
        self.seek_factor = 0
        self.arrival_factor = 1
        self.behind_distance = 0.01
        self.slowing_distance = 0.8
        self.leader_index = 0 #哪个agent是leader
        self.leader_following_factor = 5
        self.oa_factor = 15 # obstacle_avoidance_factor
        self.obstacle_radius = 1/self.window_size
        self.oa_forsee_factor = 2
        self.oa_radius = self.oa_forsee_factor*self.obstacle_radius
        
        # ORCA
        # The larger this number, the sooner an agent will respond to the presence of other agents, but the less freedom the agent has in choosing its velocities. Must be positive.
        self.timeHorizon = 20 # 10 The default minimal amount of time for which a new agent's velocities that are computed by the simulation are safe with respect to other agents. 
        self.timeHorizonObst = 10 #5 The default minimal amount of time for which a new agent's velocities that are computed by the simulation are safe with respect to obstacles.
        self.ORCA_constraints = self.max_neighbour + self.max_neighbour + 2
        self.invTimeStep = 20 # to avoid in how many time step
        self.ORCA_radius = 5 # 1.5 influence disk of 2 people added
        self.obstacle_orcaradius = 5

        self.dense_sense = 1 # 1代表使用密度滤波器
        self.dense_theta = 1
        self.dense_k = 1
        self.dense_p = 1
        # TODO: obstacle mode=2的读图处理障碍物需要补写
        # debug：ORCA障碍物还是有问题 主要是参数非常需要调整 agent pos和障碍物pos不能直接按窗口大小扩展 也不能直接存归一化后的坐标 所以有expand_size
        # 否则ORCA 障碍物distSqLine和obstacle_line.point就不准确，所以调参数还是挺麻烦的，sim_step也不能太大，否则速度改变过于突然；太小人群就运动的太慢；还是参数问题
        # TODO: 凸多边形的读图和A*寻路;kd-tree
        # 不读图片时用于设置障碍物的参数 目前只支持通过画线段来设置

        self.obstacle_setting = 0 # 0表示不存在obstacle 1要不用线段-网格模式，要不读图 2-采用凸多边形模式
        self.max_obstacle = 60 # maximum obstacle vertices
        self.resolution = self.window_size+20 #一条障碍线段会被分成多少个障碍点 改邻域搜索后这个值比window_size大其实就可以
        self.obstacles = []#[[0, 0, 1, 0],[0, 1, 1, 1],[1,0,1,1],[0,0,0,1],[0.5,0,0.5,0.45],[0.5,1,0.5,0.55]] # 线段(startx, starty, endx, endy)
        self.obstacles_x = []
        self.obstacles_y = []

        # render
        self.people_radius = 3
        self.draw_triangle = 0  # 是否在代表人的圆点中画受力三角形
        self.triangle_size = self.people_radius / self.WINDOW_WIDTH 

        self.N = 200  # people number
        self.max_speed_factor = 0.4   
        self.vel_0 = np.zeros(dtype=np.float32, shape=(self.N, 2))
        self.pos_0 = np.zeros(dtype=np.float32,shape=(self.N, 2))
        self.desiredpos_0 = np.zeros(dtype=np.float32, shape=(self.N, 2))
        self.desiredvel = np.zeros(dtype=np.float32, shape=self.N)  # 期望速度大小m/s        

        self.group_target = np.zeros(shape=(self.N, 2), dtype=np.float32)
        self.group_vel = np.zeros(shape=(self.N, 2), dtype=np.float32) #默认每个人的target 初速度都不同
        
        self.group = [] #每批的人数
        self.batch = self.N
        # social force plus
        self.gender = [] #0表示男性 1表示女性

        # state machine for steering
        self.state_machine = 1 # 1表示启用状态机
        self.state_num = 2
        self.state_target = [self.group_target,np.array([[0.5, 0.5],[0.5, 0.5]])]
        self.state_velfactor = [0.4,0.8]
          
    def post_init(self):
        """init parameters after finishing reading all modifiable parameters"""
        self.ORCA_constraints = self.max_neighbour + self.max_neighbour + 2
        self.obstacle_radius = 1/self.window_size
        self.oa_radius = self.oa_forsee_factor*self.obstacle_radius
        self.fill_desiredvel()
        if self.state_machine == 0:
            self.state_num = 1

    def fill_gender(self):
        """fixed now, can't be modified 第一批都是男性，第二批都是女性"""
        self.gender = []
        for i in range (self.batch): 
            for _ in range (self.group[i]):
                if i == 0:
                    self.gender.append(0)
                else:
                    self.gender.append(1)

    def read_img(self):
        """
        read img and fill self.im
        the path of img is self.path, must be assigned before calling this function
        """          
        im = np.array(Image.open(self.path).convert('L'))
        im = np.rot90(im,-1) # PIL读入的图片坐标系和taichi不同 需要逆时针旋转90°
        self.WINDOW_WIDTH = im.shape[0]
        self.WINDOW_HEIGHT = im.shape[1]
        self.im = ti.field(ti.f32, shape=im.shape)
        self.im.from_numpy(im)
        print("reading pic from {}, pic resolution in pixel is {}*{}".format(self.path,self.WINDOW_WIDTH,self.WINDOW_HEIGHT))

    def init_obstacle_line(self):
        """
        fill self.obstacle_x and self.obstacle_y
        so that when obstacle_setting = 2(init with convex mode)and init without picture,
        GUI can show the obstacle line
        """
        self.obstacles_x = []
        self.obstacles_y = []
        for items in self.obstacles:
            for i in range(len(items)-1):
                self.obstacles_x.append([items[i][0],items[i][1]])
                self.obstacles_y.append([items[i+1][0],items[i+1][1]])
            self.obstacles_x.append(items[len(items)-1])
            self.obstacles_y.append(items[0])

        self.obstacles_x = np.array(self.obstacles_x)
        self.obstacles_y = np.array(self.obstacles_y)

    def fill_group(self):
        """根据N总人数和group_target填充batch,group 即每批人群有多少人"""
        self.group = []
        self.batch = self.group_target.shape[0] #有几批人
        sum = 0
        batch = math.floor(self.N/self.batch)
        for _ in range (self.batch-1):
            self.group.append(batch)
            sum+=batch
        self.group.append(self.N-sum)

    def fill_desiredvel(self):
        """设置每人的期望速度 SFM的拓展 期望速度*max_speed_factor = 此人的最大速度 通过max_speed_factor来调整人群的速度"""
        for i in range(self.N):
            self.desiredvel[i] = (60+random.randint(0, 15))/100

    def fill_pos0_2batch(self,y,x1,x2):
        """调整pos_0以设置人群的初始位置,2 batch of people 必须先设置group"""        
        for i in range(self.N):
            self.pos_0[i][1] = y + random.uniform(-0.05, 0.05)
            if i < self.group[0]:
                self.pos_0[i][0] = x1
            else:
                self.pos_0[i][0] = x2
    
    def fill_pos0_circle(self):
        #circle
        for i in range(self.N):
            self.pos_0[i][0] = math.cos(2*math.pi/self.N*i)*0.4+0.5
            self.pos_0[i][1] = math.sin(2*math.pi/self.N*i)*0.4+0.5
        
    def fill_vel_target(self):
        """工具函数 按照self.group_target和self.group_vel设定每批人的目的地和初始速度 必须先设置好group group_target group_vel"""
        _vels = []
        _desiredpos = []
        for i in range (self.batch):
            for _ in range (self.group[i]):
                _vels.append(self.group_vel[i])
                _desiredpos.append(self.group_target[i])
        self.vel_0 = np.vstack(np.array(_vels))
        self.desiredpos_0 = np.vstack(np.array(_desiredpos))

    def set_scene(self,value):
        # set scene, value is the name of the scene
        if value == "circle":
            self.group_target = np.zeros(shape=(self.N, 2), dtype=np.float32)
            self.group_vel = np.zeros(shape=(self.N, 2), dtype=np.float32) #默认每个人的target 初速度都不同

            self.fill_pos0_circle()
            for i in range(self.N):
                self.group_target[i] = 1-self.pos_0[i]
                self.group_vel[i] = [0.01,0]
            self.fill_group()
            self.fill_vel_target()
        elif value == "random":
            self.group_target = np.zeros(shape=(self.N, 2), dtype=np.float32)
            self.group_vel = np.zeros(shape=(self.N, 2), dtype=np.float32) #默认每个人的target 初速度都不同
            rng = default_rng(seed=42)
            self.pos_0 = rng.random(size=(self.N, 2), dtype=np.float32)
            for i in range(self.N):
                self.group_target[i] = [0.5,0.5]
                self.group_vel[i] = [0.01,0]
            self.fill_group()
            self.fill_vel_target()
        elif value == "maze": # 右下角
            self.group_target = np.array([[0.1, 0.5]]) #每批人的目的地
            self.group_vel = np.array([[-0.01, 0.]]) #每批人的初始速度
            self.fill_group()
            self.fill_vel_target()
            self.fill_pos0_2batch(0.1,0.9,0.9)
        elif value == "long_hall":
            self.obstacles = [[0, 0, 1, 0],[0, 1, 1, 1],[1,0,1,1],[0,0,0,1],[0.5,0,0.5,0.45],[0.5,1,0.5,0.55]] # 线段(startx, starty, endx, endy)
            self.obstacles_x = np.array(self.obstacles)[:,0:2]
            self.obstacles_y = np.array(self.obstacles)[:,2:4]
            self.group_target = np.array([[0.1, 0.5],[0.9, 0.5]]) #每批人的目的地
            self.group_vel = np.array([[-0.01, 0.],[0.01,0]]) #每批人的初始速度
            self.fill_group()
            self.fill_vel_target()
            self.fill_pos0_2batch(0.5,0.9,0.1)  
        elif value == "long_hall_gendergroup":
            self.obstacles = [[0, 0, 1, 0],[0, 1, 1, 1],[1,0,1,1],[0,0,0,1],[0.5,0,0.5,0.45],[0.5,1,0.5,0.55]] # 线段(startx, starty, endx, endy)
            self.obstacles_x = np.array(self.obstacles)[:,0:2]
            self.obstacles_y = np.array(self.obstacles)[:,2:4]
            self.group_target = np.array([[0.1, 0.5],[0.9, 0.5]]) #每批人的目的地
            self.group_vel = np.array([[-0.01, 0.],[0.01,0]]) #每批人的初始速度
            self.fill_group()
            self.fill_vel_target()
            self.fill_pos0_2batch(0.5,0.9,0.1)   
            self.fill_gender()      
        elif value == "block":
            self.group_target = np.array([[0.5, 0.5],[0.5, 0.5]]) #每批人的目的地
            self.group_vel = np.array([[-0.01, 0.],[0.01,0]]) #每批人的初始速度
            self.batch = self.group_target.shape[0] #有几批人
            self.fill_group()
            self.fill_vel_target()
            self.fill_pos0_2batch(0.9,0.9,0.1)
            self.obstacles = []
            if self.obstacle_setting == 2:
                # convex obstacle
                obstacle1 = []
                obstacle1.append(utils.vec2([0.4,0.75]))
                obstacle1.append(utils.vec2([0.2, 0.75]))
                obstacle1.append(utils.vec2([0.2, 0.55]))
                obstacle1.append(utils.vec2([0.4, 0.55]))
                self.obstacles.append(obstacle1)
                obstacle1 = []
                obstacle1.append(utils.vec2([0.6,0.75]))
                obstacle1.append(utils.vec2([0.6, 0.55]))
                obstacle1.append(utils.vec2([0.8, 0.55]))
                obstacle1.append(utils.vec2([0.8, 0.75]))
                self.obstacles.append(obstacle1)
                obstacle1 = []
                obstacle1.append(utils.vec2([0.2,0.45]))
                obstacle1.append(utils.vec2([0.4, 0.45]))
                obstacle1.append(utils.vec2([0.4, 0.25]))
                obstacle1.append(utils.vec2([0.2, 0.25]))
                self.obstacles.append(obstacle1)
                obstacle1 = []
                obstacle1.append(utils.vec2([0.6,0.45]))
                obstacle1.append(utils.vec2([0.8, 0.45]))
                obstacle1.append(utils.vec2([0.8, 0.25]))
                obstacle1.append(utils.vec2([0.6, 0.25]))
                self.obstacles.append(obstacle1)
                self.init_obstacle_line()
            
    def set(self,key,value):
        # set the value of a parameter, key is the name of the parameter, value is the value of the parameter
        if key == "n":
            self.N = int(value)
            self.vel_0 = np.zeros(dtype=np.float32, shape=(self.N, 2))
            self.pos_0 = np.zeros(dtype=np.float32,shape=(self.N, 2))
            self.desiredpos_0 = np.zeros(dtype=np.float32, shape=(self.N, 2))
            self.desiredvel = np.zeros(dtype=np.float32, shape=self.N)  # 期望速度大小m/s        

            self.group_target = np.zeros(shape=(self.N, 2), dtype=np.float32)
            self.group_vel = np.zeros(shape=(self.N, 2), dtype=np.float32) #默认每个人的target 初速度都不同

            self.batch = self.N
        elif key == "max_speed_factor":
            self.max_speed_factor = float(value)
        elif key == "draw_triangle":
            self.draw_triangle = int(value)
        elif key == "triangle_size":
            self.triangle_size = float(value)
        elif key == "people_radius":
            self.people_radius = int(value)
            self.triangle_size = self.people_radius / self.WINDOW_WIDTH 
        elif key == "obstacle_setting":
            self.obstacle_setting = int(value)
        elif key == "path": # read img and set it as the background map of the crowd
            self.path = value
            self.read_img()
        elif key == "im": # clear the background map
            self.im = None
        elif key == "window_height":
            self.WINDOW_HEIGHT = int(value)
        elif key == "window_width":
            self.WINDOW_WIDTH = int(value)
        elif key == "expand_size":
            self.expand_size = int(value)
        elif key == "astar":
            self.Astar = int(value)
        elif key == "dynamic_search":
            self.dynamic_search = int(value)
        elif key == "sim_step":
            self.sim_step = float(value)
        elif key == "analysis":
            self.analysis = int(value)
        elif key == "window_size":
            self.window_size = int(value)
            self.pixel_number = self.window_size ** 2
            self.resolution = self.window_size+20
        elif key == "resolution":
            self.resolution == int(value)
        elif key == "search_radius":
            self.search_radius = int(value)
        elif key == "stop_radius":
            self.stop_radius = float(value)
        elif key == "max_neighbour":
            self.max_neighbour = int(value)
        elif key == "desired_factor":
            self.desired_factor = float(value)
        elif key == "goal_threshold":
            self.goal_threshold = float(value)
        elif key == "relaxation_time":
            self.relaxation_time = float(value)
        elif key == "timeHorizon":
            self.timeHorizon = float(value)
        elif key == "timeHorizonObst":
            self.timeHorizonObst = float(value)
        elif key == "orca_constraints":
            self.ORCA_constraints = int(value)
        elif key == "invtimestep":
            self.invTimeStep = float(value)
        elif key == "orca_radius":
            self.ORCA_radius = float(value)
        elif key == "obstacle_orcaradius":
            self.obstacle_orcaradius = float(value)
        elif key == "max_obstacle":
            self.max_obstacle = int(value)

        elif key == "self.dense_sense":
            self.dense_sense = int(value)
        elif key =="dense_theta":
            self.dense_theta = float(value)
        elif key =="dense_k":
            self.dense_k = float(value)
        elif key =="dense_p":
            self.dense_p = float(value)
        
        elif key == "use_gender":
            self.use_gender = int(value)
        elif key == "gender_gap":
            self.gender_gap = float(value)
        elif key == "social_radius":
            self.social_radius = float(value)
        elif key == "lambda_importance":
            self.lambda_importance = float(value)
        elif key == "gamma":
            self.gamma = float(value)
        elif key == "n_fij":
            self.n_fij = float(value)
        elif key == "n_prime":
            self.n_prime = float(value)
        elif key == "fij_factor":
            self.fij_factor = float(value)
        elif key == "shoulder_radius":
            self.shoulder_radius = float(value)
        elif key == "sigma":
            self.sigma = float(value)
        elif key == "obstacle_threshold":
            self.obstacle_threshold = float(value)
        elif key == "obstacle_factor":
            self.obstacle_factor = float(value)

        elif key == "flocking":
            self.flocking = int(value)
        elif key == "flocking_radius":
            self.flocking_radius = float(value)
        elif key == "alignment_factor":
            self.alignment_factor = float(value)
        elif key == "separation_factor":
            self.separation_factor = float(value)
        elif key == "cohesion_factor":
            self.cohesion_factor = float(value)

        elif key == "steering_force":
            self.steering_force = int(value)
        elif key == "flee_factor":
            self.flee_factor = float(value)
        elif key == "seek_factor":
            self.seek_factor = float(value)
        elif key == "arrival_factor":
            self.arrival_factor = float(value)
        elif key == "behind_distance":
            self.behind_distance = float(value)
        elif key == "slowing_distance":
            self.slowing_distance = float(value)
        elif key == "leader_following_factor":
            self.leader_following_factor = float(value)
        elif key =="leader_index":
            self.leader_index = int(value)
        elif key == "oa_factor":
            self.oa_factor = float(value)
        elif key == "oa_forsee_factor":
            self.oa_forsee_factor = float(value)

        elif key == "state_machine":
            self.state_machine = int(value)
        elif key == "state_num":
            self.state_num = int(value)
        elif key == "state_target":
            self.state_target = json.loads(value)
        elif key == "state_velfactor":
            self.state_velfactor = json.loads(value)